//	温度・湿度・気圧センサー BME280(I2C版 モジュール名 GYBMEP) ライブラリ
//	『昼夜逆転』工作室	http://jsdiy.starfree.jp

/*
【更新履歴】
2017/06/18	v1.00	初版
	ベースとしたAPI: BME280_driver-master [Bosch Sensortec (github)]
2017/07/27	v1.10
	メモリ節約のため、センサー値と校正値の読み込みバッファを兼用にした。
	外部からバッファを与えられるようにした。
2023/07/27	v1.10e	←ぴったり6年！偶然！
	ESP32(Arduino環境)用に一部変更：includeと通信部分とdelay()関数を変更。
	動作内容、パラメータ等は変更なし。
*/

#include <Arduino.h>
#include <Wire.h>
#include "Bme280.hpp"


//BME280モジュールのスレーブアドレス(7bit)
//#define	BME280_SLAVE_ADDRESS	0x76	//ヘッダファイルへ移動
/*
BME280のSDO端子でスレーブアドレスのLSBを決めることができる。
GND  とショートなら'0'→スレーブアドレス:0x76
VDDIOとショートなら'1'→スレーブアドレス:0x77
当モジュール'GYBMEP'では0x76となっている。
→センサー下3つのパッドは左から GND,SDO,VDDIO で、GND-SDOのパターンがつながっている。
→パターンカットしてSDO-VDDIOをハンダショートして0x77にしたり、
　GND-SDOをハンダショートして0x76に戻したりできる。

プルアップ抵抗は10kΩが当モジュールに実装済。
*/

//=========================================================
//	Calibration data
//=========================================================

//read calibration data: define
//(1) レジスタcalib00..calib25
#define CALPARAMS1_START_ADDR	0x88
#define CALPARAMS1_END_ADDR		0xA1

//(2) レジスタcalib26..calib41
#define CALPARAMS2_START_ADDR	0xE1
#define CALPARAMS2_END_ADDR		0xF0

//レジスタpress_msb..hum_lsb
#define SENSORVALUE_START_ADDR	0xF7
#define SENSORVALUE_END_ADDR	0xFE

#define BME280_BYTE_LENGTH(st,ed)	((ed) - (st) + 1)
#define BME280_MAX(a,b)	((a) < (b) ? (b) : (a))

//バッファサイズ計算
#define CALPARAMS1_LENGTH	BME280_BYTE_LENGTH(CALPARAMS1_START_ADDR, CALPARAMS1_END_ADDR)
#define CALPARAMS2_LENGTH	BME280_BYTE_LENGTH(CALPARAMS2_START_ADDR, CALPARAMS2_END_ADDR)
#define CALPARAMS_LENGTH	BME280_MAX(CALPARAMS1_LENGTH, CALPARAMS2_LENGTH)	//28byte
#define SENSORVALUE_LENGTH	BME280_BYTE_LENGTH(SENSORVALUE_START_ADDR, SENSORVALUE_END_ADDR)	//8byte
#define BME280_BUFFER_LENGTH	BME280_MAX(CALPARAMS_LENGTH, SENSORVALUE_LENGTH)	//結果28byte

//バッファ確保（press,temp,humのレジスタ読み込み／校正値読み込み、で兼用）
static	uint8_t	regBuf[BME280_BUFFER_LENGTH];	//Init()で初期化する場合
//static	uint8_t* regBuf;	//InitWithMemory()で初期化する場合	★AVRでメモリ節約していたときの名残

//バイト配列のインデックス
enum ERegBufIndex
{
	press_msb = 0,
	press_lsb,
	press_xlsb,
	temp_msb,
	temp_lsb,
	temp_xlsb,
	hum_msb,
	hum_lsb
};

//=========================================================
//	Compensation parameter
//=========================================================

//校正値を格納する構造体
typedef	struct
{
	uint16_t	t1;
	int16_t		t2;
	int16_t		t3;
	uint16_t	p1;
	int16_t		p2;
	int16_t		p3;
	int16_t		p4;
	int16_t		p5;
	int16_t		p6;
	int16_t		p7;
	int16_t		p8;
	int16_t		p9;
	uint8_t		h1;
	int16_t		h2;
	uint8_t		h3;
	int16_t		h4;
	int16_t		h5;
	int8_t		h6;
	//
	int32_t	t_fine;
}
TBme280ComParam;

//センサー読込値から表示値を算出する際の校正値
static	TBme280ComParam	dig;

//=========================================================
//	Control registers
//=========================================================

//Humidity oversampling
//Pressure oversampling
//Temperature oversampling
typedef	enum
{
	BME280_OSRS_Skipped			= 0b000,
	BME280_OSRS_Oversamplingx1	= 0b001,
	BME280_OSRS_Oversamplingx2	= 0b010,
	BME280_OSRS_Oversamplingx4	= 0b011,
	BME280_OSRS_Oversamplingx8	= 0b100,
	BME280_OSRS_Oversamplingx16	= 0b101
}
EBme280Osrs;

//sensor mode
typedef	enum
{
	BME280_MODE_SleepMode	= 0b00,
	BME280_MODE_ForcedMode	= 0b01,	//and 0b10
	BME280_MODE_NormalMode	= 0b11
}
EBme280SensorMode;

//inactive duration t_standby
typedef	enum
{
	BME280_TSB_1ms		= 0b000,	//0.5ms
	BME280_TSB_63ms		= 0b001,	//62.5ms
	BME280_TSB_125ms	= 0b010,
	BME280_TSB_250ms	= 0b011,
	BME280_TSB_500ms	= 0b100,
	BME280_TSB_1000ms	= 0b101,
	BME280_TSB_10ms		= 0b110,
	BME280_TSB_20ms		= 0b111
}
EBme280Tsb;

//time constant of the IIR filter
typedef	enum
{
	BME280_FILTER_FilterOff		= 0b000,
	BME280_FILTER_coefficient2	= 0b001,
	BME280_FILTER_coefficient4	= 0b010,
	BME280_FILTER_coefficient8	= 0b011,
	BME280_FILTER_coefficient16	= 0b100
}
EBme280Filter;

//Setting
typedef	struct
{
	//oversampling resolution
	EBme280Osrs	osrs_h;
	EBme280Osrs	osrs_p;
	EBme280Osrs	osrs_t;

	//sensor mode
	EBme280SensorMode	mode;
	
	//t_standby
	EBme280Tsb	t_sb;
	
	//filter coefficient
	EBme280Filter	filter;
}
TBme280Setting;

//センサーの目的別設定値
static	TBme280Setting settingParam;

/*	Bme280::privateへ
//=========================================================
//	プロトタイプ宣言
//=========================================================

static	void	WriteRegValue(uint8_t regAddr, uint8_t dataByte);
static	void	ReadRegValue(uint8_t startAddr, uint8_t readLength, uint8_t* retBuf);
static	void	DelayMSec(uint8_t msec);
static	uint8_t	ComCheck(void);
static	void	SoftReset(void);
static	void	ReadCalData(void);
static	void	WriteSensorParamToReg(void);
static	uint8_t	ComputeWaitTime(void);
static	uint16_t	GetModulo(int32_t num, uint16_t modorder);
static	int32_t		CompensateTemperature(int32_t ut);
static	uint32_t	CompensatePressure(uint32_t up);
static	uint32_t	CompensateHumidity(int32_t uh);
//
static	void	PresetPatternWeatherMonitoring(void);
*/

//=========================================================

//レジスタアドレスXXhにデータ0xYYを書く
void	Bme280::WriteRegValue(uint8_t regAddr, uint8_t dataByte)
{
/*
//	TWI_Start();
//	TWI_SlaveAddrW(BME280_SLAVE_ADDRESS);
	TWI_Start((BME280_SLAVE_ADDRESS << 1) | SLA_W_BIT);
	TWI_WriteByte(regAddr);
	TWI_WriteByte(dataByte);
	TWI_Stop();
*/
	Wire.beginTransmission(slaveAddress);
	Wire.write(regAddr);
	Wire.write(dataByte);
	Wire.endTransmission(true);
}

//指定アドレスに格納された値を読む
void	Bme280::ReadRegValue(uint8_t startAddr, uint8_t readLength, uint8_t* retBuf)
{
/*
	uint8_t i;
	
//	TWI_Start();
//	TWI_SlaveAddrW(BME280_SLAVE_ADDRESS);
	TWI_Start((BME280_SLAVE_ADDRESS << 1) | SLA_W_BIT);
	TWI_WriteByte(startAddr);
	TWI_Stop();
	
//	TWI_Start();
//	TWI_SlaveAddrR(BME280_SLAVE_ADDRESS);
	TWI_Start((BME280_SLAVE_ADDRESS << 1) | SLA_R_BIT);
	for (i = 0; i < readLength; i++)
	{
		TWI_ReadByte((i == readLength - 1), retBuf + i);
	}
	TWI_Stop();
*/
	Wire.beginTransmission(slaveAddress);
	Wire.write(startAddr);
	Wire.endTransmission(false);	//false:接続を維持したまま送信を完了する
	Wire.requestFrom(slaveAddress, readLength);
	for (uint8_t i = 0; i < readLength; i++)
	{
		retBuf[i] = (uint8_t)Wire.read();
	}
}

//1ms単位の時間待ち
void	Bme280::DelayMSec(uint8_t msec)
{
/*
	uint8_t i;
	for (i = 0; i < msec; i++) { _delay_ms(1); }
*/
	delay(msec);
}

//初期化
void	Bme280::BME280_Init(void)
{
	/*
	if (ComCheck()) { ReadCalData(); }
	*/
	if (ComCheck())
	{
		ReadCalData();
		PresetPatternWeatherMonitoring();	//プリセットパターンを'WeatherMonitoring'固定で呼び出す
	}
}

/*	★AVRでメモリ節約していたときの名残
//初期化
//・メモリ節約のため他モジュールから作業用バッファを与える。
void	Bme280::BME280_InitWithMemory(uint8_t* buffer)
{
	regBuf = buffer;
	if (ComCheck())
	{
		ReadCalData();
		PresetPatternWeatherMonitoring();	//プリセットパターンを'WeatherMonitoring'固定で呼び出す
	}
}
*/

//通信確認
uint8_t	Bme280::ComCheck(void)
{
	//レジスタアドレスD0hを読んで0x60だったら通信できていることになる
	uint8_t	data;
	ReadRegValue(0xD0, 1, &data);
	
	return (data == 0x60);
}

//ソフトリセット
void	Bme280::SoftReset(void)
{
	//レジスタアドレスE0hに0xB6を書く
	WriteRegValue(0xE0, 0xB6);

	//「Start-up time: Max 2ms」
	//というわけでハードリセットと同じ時間だけ待つ（さらに余裕を持たせた）
	delay(10);
}

//測定準備状態を取得する
#if (0)	//特に必要ない
void	BME280_GetStatus(uint8_t* measuring, uint8_t* im_update)
{
	//レジスタアドレスF3hを読む
	uint8_t	status;
	ReadRegValue(0xF3, 1, &status);

	*measuring = (status >> 3) & 0x01;	//bit3
	*im_update = (status >> 0) & 0x01;	//bit0

	/*
	measuring:
	Automatically set to ‘1’ whenever a conversion is running
	and back to ‘0’ when the results have been transferred to the data registers.
	
	im_update:
	Automatically set to ‘1’ when the NVM data are being copied to image registers
	and back to ‘0’ when the copying is done.
	The data are copied at power-on-reset and before every conversion.
	*/
}
#endif

//read calibration data
void	Bme280::ReadCalData(void)
{
	uint8_t* buf = regBuf;	//短い名前に読み替え
	uint8_t i;
	
	//calibration dataを読む(1)
	ReadRegValue(CALPARAMS1_START_ADDR, CALPARAMS1_LENGTH, buf);

	//バッファの値を加工して構造体へ格納する
	i = 0;
	dig.t1 =			(((uint16_t)buf[i]) | ((uint16_t)buf[i + 1] << 8));		i += 2;
	dig.t2 =	(int16_t)(((uint16_t)buf[i]) | ((uint16_t)buf[i + 1] << 8));	i += 2;
	dig.t3 =	(int16_t)(((uint16_t)buf[i]) | ((uint16_t)buf[i + 1] << 8));	i += 2;
	dig.p1 =			(((uint16_t)buf[i]) | ((uint16_t)buf[i + 1] << 8));		i += 2;
	dig.p2 =	(int16_t)(((uint16_t)buf[i]) | ((uint16_t)buf[i + 1] << 8));	i += 2;
	dig.p3 =	(int16_t)(((uint16_t)buf[i]) | ((uint16_t)buf[i + 1] << 8));	i += 2;
	dig.p4 =	(int16_t)(((uint16_t)buf[i]) | ((uint16_t)buf[i + 1] << 8));	i += 2;
	dig.p5 =	(int16_t)(((uint16_t)buf[i]) | ((uint16_t)buf[i + 1] << 8));	i += 2;
	dig.p6 =	(int16_t)(((uint16_t)buf[i]) | ((uint16_t)buf[i + 1] << 8));	i += 2;
	dig.p7 =	(int16_t)(((uint16_t)buf[i]) | ((uint16_t)buf[i + 1] << 8));	i += 2;
	dig.p8 =	(int16_t)(((uint16_t)buf[i]) | ((uint16_t)buf[i + 1] << 8));	i += 2;
	dig.p9 =	(int16_t)(((uint16_t)buf[i]) | ((uint16_t)buf[i + 1] << 8));	i += 2;
	dig.h1 =	(uint8_t)buf[i];

	//calibration dataを読む(2)
	ReadRegValue(CALPARAMS2_START_ADDR, CALPARAMS2_LENGTH, buf);

	//バッファの値を加工して構造体へ格納する
	i = 0;
	dig.h2 =	(int16_t)(((uint16_t)buf[i]) | ((uint16_t)buf[i + 1] << 8));	i += 2;
	dig.h3 =	(uint8_t)buf[i];	i++;
	dig.h4 =	(int16_t)(((uint16_t)buf[i] << 4) | ((uint16_t)(buf[i + 1] & 0x0F)));
	dig.h5 =	(int16_t)(((uint16_t)((buf[i + 1] & 0xF0) >> 4)) | ((uint16_t)buf[i + 2] << 4));	i += 3;
	dig.h6 =	(int8_t)buf[i];
}

//データシートに掲載されているお勧めの測定パターンをセットする
/*	datasheet:
	5.5 Recommended modes of operation
	5.5.1 Weather monitoring
	5.5.2 Humidity sensing
	5.5.3 Indoor navigation
	5.5.4 Gaming
*/
/*
void	BME280_PresetPattern(EBme280Pattern pattern)
{
	switch (pattern)
	{
	case BME280_PATTERN_WeatherMonitoring:
		settingParam.mode = BME280_MODE_ForcedMode;	//1 sample / minute
		settingParam.osrs_p = BME280_OSRS_Oversamplingx1;
		settingParam.osrs_t = BME280_OSRS_Oversamplingx1;
		settingParam.osrs_h = BME280_OSRS_Oversamplingx1;
		settingParam.filter = BME280_FILTER_FilterOff;
		break;

	case BME280_PATTERN_HumiditySensing:
		settingParam.mode = BME280_MODE_ForcedMode;	//1 sample / second
		settingParam.osrs_p = BME280_OSRS_Skipped;
		settingParam.osrs_t = BME280_OSRS_Oversamplingx1;
		settingParam.osrs_h = BME280_OSRS_Oversamplingx1;
		settingParam.filter = BME280_FILTER_FilterOff;
		break;

	case BME280_PATTERN_IndoorNavigation:
		settingParam.mode = BME280_MODE_NormalMode;
		settingParam.t_sb = BME280_TSB_1ms;		//T_standby = 0.5ms
		settingParam.osrs_p = BME280_OSRS_Oversamplingx16;
		settingParam.osrs_t = BME280_OSRS_Oversamplingx2;
		settingParam.osrs_h = BME280_OSRS_Oversamplingx1;
		settingParam.filter = BME280_FILTER_coefficient16;
		break;

	case BME280_PATTERN_Gaming:
		settingParam.mode = BME280_MODE_NormalMode;
		settingParam.t_sb = BME280_TSB_1ms;		//T_standby = 0.5ms
		settingParam.osrs_p = BME280_OSRS_Oversamplingx4;
		settingParam.osrs_t = BME280_OSRS_Oversamplingx1;
		settingParam.osrs_h = BME280_OSRS_Skipped;
		settingParam.filter = BME280_FILTER_coefficient16;
		break;

	default:
		break;
	}

	//レジスタに測定パラメータを書き込む
	SoftReset();
	WriteSensorParamToReg();
}
*/
void	Bme280::PresetPatternWeatherMonitoring(void)
{
	//BME280_PATTERN_WeatherMonitoring
	settingParam.mode = BME280_MODE_ForcedMode;	//1 sample / minute
	settingParam.osrs_p = BME280_OSRS_Oversamplingx1;
	settingParam.osrs_t = BME280_OSRS_Oversamplingx1;
	settingParam.osrs_h = BME280_OSRS_Oversamplingx1;
	settingParam.filter = BME280_FILTER_FilterOff;
	
	//レジスタに測定パラメータを書き込む
	SoftReset();
	WriteSensorParamToReg();
}

//センサー動作の設定値をレジスタに書き込む
void	Bme280::WriteSensorParamToReg(void)
{
	uint8_t	configRegValue, ctrlMeasRegValue, ctrlHumRegValue;
	uint8_t msec;

	configRegValue = ((settingParam.t_sb << 5) | (settingParam.filter << 2)) & 0xFC;	//LSB側2bitは00とする
	WriteRegValue(0xF5, configRegValue);
	
	ctrlHumRegValue = settingParam.osrs_h & 0x07;
	WriteRegValue(0xF2, ctrlHumRegValue);	//ctrlMeasRegValueより前に書き込むこと
	
	ctrlMeasRegValue = ((settingParam.osrs_t << 5) | (settingParam.osrs_p << 2) | settingParam.mode) & 0xFF;
	WriteRegValue(0xF4, ctrlMeasRegValue);	//ctrlHumRegValueより後に書き込むこと

	//測定パラメータのレジスタを書き込んだ後、測定データを読み込めるまでの時間待ち
	msec = ComputeWaitTime();
	DelayMSec(msec);
}

//DELAY DEFINITIONS
#define	T_INIT_MAX				(20)	// 20/16 = 1.25 ms
#define	T_MEASURE_PER_OSRS_MAX	(37)	// 37/16 = 2.3125 ms
#define	T_SETUP_PRESSURE_MAX	(10)	// 10/16 = 0.625 ms
#define	T_SETUP_HUMIDITY_MAX	(10)	// 10/16 = 0.625 ms

//Computing waiting time for sensor data read
//Return value: The value of delay time for force mode
uint8_t	Bme280::ComputeWaitTime()
{
	uint8_t	msec;

	msec = (T_INIT_MAX + T_MEASURE_PER_OSRS_MAX *
		(((1 << settingParam.osrs_t) >> 1) +
		((1 << settingParam.osrs_p) >> 1) +
		((1 << settingParam.osrs_h) >> 1)) +
		((settingParam.osrs_p > 0) ? T_SETUP_PRESSURE_MAX : 0) +
		((settingParam.osrs_h > 0) ? T_SETUP_HUMIDITY_MAX : 0) + 15) / 16;
	
	return msec;
}

//mode値を取得する
#if (0)	//未使用
static	EBme280SensorMode	GetSensorMode()
{
	uint8_t	ctrlMeasReg, mode;
	EBme280SensorMode modes[] =
	{
		BME280_MODE_SleepMode,
		BME280_MODE_ForcedMode, BME280_MODE_ForcedMode,
		BME280_MODE_NormalMode
	};

	ReadRegValue(0xF4, 1, &ctrlMeasReg);
	mode = ctrlMeasReg & 0x03;
	return modes[mode];
}
#endif

//センサーから測定値を取得する（表示値に変換済)
void	Bme280::BME280_GetSensorValue(TBme280Value* retValue)
{
	int32_t ut, ct;
	uint32_t up, cp, uh, ch;
	uint8_t* buf = regBuf;	//短い名前に読み替え

	//ForcedModeの場合、modeレジスタにForcedModeを書き込む（SleepMode状態から測定状態へ遷移させる）
	if (settingParam.mode == BME280_MODE_ForcedMode) { WriteSensorParamToReg(); }

	//read sensor value - Press, Temp, Hum
	ReadRegValue(SENSORVALUE_START_ADDR, SENSORVALUE_LENGTH, buf);

	//uncompensated {pressure, temperature, humidity} value
	up =		((uint32_t)buf[press_msb] << 12) | ((uint32_t)buf[press_lsb] << 4) | ((uint32_t)buf[press_xlsb] >> 4);
	ut = (int32_t)(((uint32_t)buf[temp_msb] << 12) | ((uint32_t)buf[temp_lsb] << 4) | ((uint32_t)buf[temp_xlsb] >> 4));
	uh =		((uint32_t)buf[hum_msb] << 8) | (uint32_t)buf[hum_lsb];

	//表示値を算出する
	ct = CompensateTemperature(ut);
	retValue->temperature.integer = (int16_t)(ct / 100);
	retValue->temperature.decimal = GetModulo(ct);	//小数第2位までの値 → (0.)00～(0.)99
	retValue->temperature.decimal /= 10;	//小数第1位までとする

	cp = CompensatePressure(up);
	retValue->pressure.integer = (int16_t)(cp / 100);
	retValue->pressure.decimal = (uint8_t)(cp % 100);	//小数第2位までの値 → (0.)00～(0.)99
	retValue->pressure.decimal /= 10;	//小数第1位までとする
	
	ch = CompensateHumidity(uh);
	retValue->humidity.integer = (int16_t)(ch / 1024);
	uint16_t huDecimal = (uint16_t)(((ch % 1024) * 1000) / 1024);	//小数第3位までの値(10進数) → (0.)000～(0.)999
	retValue->humidity.decimal = (uint8_t)(huDecimal / 100);	//小数第1位までとする
}

//剰余を求める（負数に対応）
uint8_t	Bme280::GetModulo(int32_t num)
{
	uint32_t unum = (num < 0) ? -num : num;
	return (uint8_t)(unum % 100);
}

//Reads actual temperature from uncompensated temperature
//Output value of "5123" equals 51.23 DegC.
int32_t	Bme280::CompensateTemperature(int32_t ut)
{
	int32_t	x1, x2;
	int32_t	temperature;

	// calculate x1
	x1 = ((((ut >> 3) - ((int32_t)dig.t1 << 1))) * ((int32_t)dig.t2)) >> 11;
	
	// calculate x2
	x2 = (((((ut >> 4) - ((int32_t)dig.t1)) * ((ut >> 4) - ((int32_t)dig.t1))) >> 12) * ((int32_t)dig.t3)) >> 14;
	
	// calculate t_fine
	//fine resolution temperature value over to the pressure compensation formula
	dig.t_fine = x1 + x2;

	// calculate temperature
	temperature = (dig.t_fine * 5 + 128) >> 8;
	return	temperature;
}

//Reads actual pressure from uncompensated pressure
//Output value of "96386" equals 96386 Pa = 963.86 hPa = 963.86 millibar
#define	BME280_INIT_VALUE	(0)
#define	BME280_INVALID_DATA	(0)

uint32_t	Bme280::CompensatePressure(uint32_t up)
{
	int32_t x1 = BME280_INIT_VALUE;
	int32_t x2;
	uint32_t pressure;

	// calculate x1
	x1 = (((int32_t)dig.t_fine) >> 1) - (int32_t)64000;
	
	// calculate x2
	x2 = (((x1 >> 2) * (x1 >> 2)) >> 11) * ((int32_t)dig.p6);
	x2 = x2 + ((x1 * ((int32_t)dig.p5)) << 1);
	x2 = (x2 >> 2) + (((int32_t)dig.p4) << 16);
	
	// calculate x1
	x1 = (((dig.p3 * (((x1 >> 2) * (x1 >> 2)) >> 13)) >> 3) + ((((int32_t)dig.p2) * x1) >> 1)) >> 18;
	x1 = ((((32768 + x1)) * ((int32_t)dig.p1)) >> 15);

	// calculate pressure
	pressure = (((uint32_t)(((int32_t)1048576) - up) - (x2 >> 12))) * 3125;
	
	if (pressure < 0x80000000) {
		// Avoid exception caused by division by zero
		if (x1 != BME280_INIT_VALUE) {
			pressure = (pressure << 1) / ((uint32_t)x1);
		} else {
			return BME280_INVALID_DATA;
		}
	} else {
		// Avoid exception caused by division by zero
		if (x1 != BME280_INIT_VALUE) {
			pressure = (pressure / (uint32_t)x1) * 2;
		} else {
			return BME280_INVALID_DATA;
		}
	}

	x1 = (((int32_t)dig.p9) * ((int32_t)(((pressure >> 3) * (pressure >> 3)) >> 13))) >> 12;
	x2 = (((int32_t)(pressure >> 2)) * ((int32_t)dig.p8)) >> 13;
	
	pressure = (uint32_t)((int32_t)pressure + ((x1 + x2 + dig.p7) >> 4));
	return pressure;
}

//Reads actual humidity from uncompensated humidity
//Returns the value in %rH as unsigned 32bit integer in Q22.10 format(22 integer 10 fractional bits).
//An output value of 42313 represents 42313 / 1024 = 41.321 %rH
uint32_t	Bme280::CompensateHumidity(int32_t uh)
{
	int32_t x1;
	uint32_t humidity;

	// calculate x1
	x1 = (dig.t_fine - ((int32_t)76800));
	x1 = (((((uh << 14) - (((int32_t)dig.h4) << 20) - (((int32_t)dig.h5) * x1)) +
		((int32_t)16384)) >> 15) * (((((((x1 * ((int32_t)dig.h6)) >> 10) *
		(((x1 * ((int32_t)dig.h3)) >> 11) + ((int32_t)32768))) >> 10) +
		((int32_t)2097152)) * ((int32_t)dig.h2) + 8192) >> 14));
	x1 = (x1 - (((((x1 >> 15) * (x1 >> 15)) >> 7) * ((int32_t)dig.h1)) >> 4));
	x1 = (x1 < 0 ? 0 : x1);
	x1 = (x1 > 419430400 ? 419430400 : x1);
	
	humidity = (uint32_t)(x1 >> 12);
	return humidity;
}
