/*
Example Code To Get ESP32 To Connect To A Router Using WPS
===========================================================
This example code provides both Push Button method and Pin
based WPS entry to get your ESP connected to your WiFi router.

Hardware Requirements
========================
ESP32 and a Router having WPS functionality

This code is under Public Domain License.

Author:
Pranav Cherukupalli <cherukupallip@gmail.com>
*/

#include <Arduino.h>
#include "WiFiWps.hpp"

/*
Change the definition of the WPS mode
from WPS_TYPE_PBC to WPS_TYPE_PIN in
the case that you are using pin type
WPS
*/
#define ESP_WPS_MODE      WPS_TYPE_PBC
#define ESP_MANUFACTURER  "ESPRESSIF"
#define ESP_MODEL_NUMBER  "ESP32"
#define ESP_MODEL_NAME    "ESPRESSIF IOT"
#define ESP_DEVICE_NAME   "ESP STATION"

static	esp_wps_config_t	config;
static	void	wpsInitConfig(void);
static	void	wpsStart(void);
static	void	wpsStop(void);
static	String	wpspin2string(uint8_t a[]);
static	void	WiFiEvent(WiFiEvent_t event, arduino_event_info_t info);
static	void	startingWps(void);

static	void	wpsInitConfig(void)
{
  config.wps_type = ESP_WPS_MODE;
  strcpy(config.factory_info.manufacturer, ESP_MANUFACTURER);
  strcpy(config.factory_info.model_number, ESP_MODEL_NUMBER);
  strcpy(config.factory_info.model_name, ESP_MODEL_NAME);
  strcpy(config.factory_info.device_name, ESP_DEVICE_NAME);
}

static	void	wpsStart(void)
{
	if(esp_wifi_wps_enable(&config)){
		Serial.println("WPS Enable Failed");
	} else if(esp_wifi_wps_start(0)){
		Serial.println("WPS Start Failed");
	}
}

static	void	wpsStop(void)
{
	if(esp_wifi_wps_disable()){
		Serial.println("WPS Disable Failed");
	}
}

static	String	wpspin2string(uint8_t a[])
{
  char wps_pin[9];
  for(int i=0;i<8;i++){
	wps_pin[i] = a[i];
  }
  wps_pin[8] = '\0';
  return (String)wps_pin;
}

static	void	WiFiEvent(WiFiEvent_t event, arduino_event_info_t info)
{
  switch(event){
	case ARDUINO_EVENT_WIFI_STA_START:
	  Serial.println("Station Mode Started");
	  break;
	case ARDUINO_EVENT_WIFI_STA_GOT_IP:
	  Serial.println("Connected to :" + String(WiFi.SSID()));
	  Serial.print("Got IP: ");
	  Serial.println(WiFi.localIP());
	  break;
	case ARDUINO_EVENT_WIFI_STA_DISCONNECTED:
	  Serial.println("Disconnected from station, attempting reconnection");
	  WiFi.reconnect();
	  break;
	case ARDUINO_EVENT_WPS_ER_SUCCESS:
	  Serial.println("WPS Successfull, stopping WPS and connecting to: " + String(WiFi.SSID()));
	  wpsStop();
	  delay(10);
	  WiFi.begin();
	  break;
	case ARDUINO_EVENT_WPS_ER_FAILED:
	  Serial.println("WPS Failed, retrying");
	  wpsStop();
	  wpsStart();
	  break;
	case ARDUINO_EVENT_WPS_ER_TIMEOUT:
	  Serial.println("WPS Timedout, retrying");
	  wpsStop();
	  wpsStart();
	  break;
	case ARDUINO_EVENT_WPS_ER_PIN:
	  Serial.println("WPS_PIN = " + wpspin2string(info.wps_er_pin.pin_code));
	  break;
	default:
	  break;
  }
}

//WPSで接続開始する
static	void	startingWps(void)
{
  WiFi.onEvent(WiFiEvent);
  WiFi.mode(WIFI_MODE_STA);
  Serial.println("Starting WPS");
  wpsInitConfig();
  wpsStart();
}

//main.cppのsetup()から呼び出す
bool	WiFiWps::wpsSetup(void)
{
	WiFi.mode(WIFI_MODE_STA);
	WiFi.begin();	//WPSの場合は引数(ssid, passphrase)は不要

	uint8_t count = 0;
	while (WiFi.status() != WL_CONNECTED)
	{
		delay(1000);	//1秒
//Serial.print(".");

		if (++count == 5)
		{
//Serial.println("WiFi Timedout.");
			return false;
		}
	}

	//WiFi接続成功
//Serial.println("WiFi connected.");
//Serial.println("IP address: " + WiFi.localIP().toString());
	return true;
}

//WiFiを切断する（ライト／ディープスリープ開始のため）
//・ライト／ディープスリープ前にWiFiを切断しないと、スリープ復帰後にリセットがかかる。
bool	WiFiWps::wpsDisconnect(void)
{
	return WiFi.disconnect(true);
}

//切断したWiFiを接続し直す（ライト／ディープスリープ復帰後のWiFi再開のため）
bool	WiFiWps::wpsReconnect(void)
{
	return WiFi.reconnect();
}
