//	BME280/VFD使用 時計兼環境センサー表示器
//	『昼夜逆転』工作室	http://jsdiy.starfree.jp
//	2023/09	新規作成

#include <Arduino.h>
#include "Record.hpp"

//センサー値をファイルに書き込む
void	Record::write(const struct tm& timeInfo, const TBme280Value& sensorValue)
{
	//保存ファイルが利用可能な状態であることを確認する
	fileAvailable();

	//レコードを保存する
	String record = createRecord(timeInfo, sensorValue);
	bool isOK = update(record);
Serial.printf("file update is %s.\n", isOK ? "OK" : "NG");
	if (!isOK) { return; }

	//タイムスタンプのファイルを更新する
	lastUpdateTimestamp = buildTimestamp(timeInfo);
	writeTimestamp(lastUpdateTimestamp);
}

//今回の書き込もうとしている時間(Hour)が前回の時間(Hour)と同じか（記録済みの時間(Hour)か）
//・戻り値	true:この時間(Hour)は記録済み、またはエラー発生,	false:未記録
bool	Record::isUpdatedThisHour(const struct tm& timeInfo)
{
	String nowTime = buildTimestamp(timeInfo);
	return (0 < nowTime.length()) && nowTime.equals(lastUpdateTimestamp);
}

//レコードの最終保存時間を読み込む
String	Record::readTimestamp(void)
{
	if (!SPIFFS.exists(fileUpdateTime)) { return ""; }
	File file = SPIFFS.open(fileUpdateTime, FILE_READ);
	if (!file) { return ""; }
	String timestamp = file.readString();
	file.close();
	timestamp.trim();
	return timestamp;
}

//レコードの最終保存時間を書き込む
bool	Record::writeTimestamp(const String& timestamp)
{
	return writeFile(fileUpdateTime, timestamp);
}

//ファイルに書き込む
//・上書き。ファイルがなければ作成して書込み。
bool	Record::writeFile(const String& filename, const String& data)
{
	File file = SPIFFS.open(filename, FILE_WRITE, true);
	if (!file) { return false; }
	size_t length = file.print(data);
	file.close();
	return (length == data.length());
}

//タイムスタンプを作成する
String	Record::buildTimestamp(const struct tm& timeInfo)
{
	return String(timeInfo.tm_yday + timeInfo.tm_hour);
}

//レコードを作成する
String	Record::createRecord(const struct tm& tTimeInfo, const TBme280Value& tSensorValue)
{
	//CSVのフォーマット："yyyy/mm/dd","hh:mm","tt.t","hh.h","pppp.p"	//年/月/日,時:分,温度,湿度,気圧
	char buf[100];
	const char* fmt = R"**("%d/%d/%d","%d:%02d","%d.%d","%d.%d","%d.%d")**";

	int	length = sprintf(buf, fmt,
		tTimeInfo.tm_year + 1900, tTimeInfo.tm_mon + 1, tTimeInfo.tm_mday,
		tTimeInfo.tm_hour, tTimeInfo.tm_min,
		tSensorValue.temperature.integer, tSensorValue.temperature.decimal,
		tSensorValue.humidity.integer, tSensorValue.humidity.decimal,
		tSensorValue.pressure.integer, tSensorValue.pressure.decimal);
Serial.printf("create record:");	Serial.println(buf);
	return String(buf, length);
}

//csvファイルのみある、またはcsvファイルとbakファイルがある状態（正常な状態）に整える
void	Record::fileAvailable(void)
{
	if (SPIFFS.exists(fileCsv))
	{
		//問題ない状態
	}
	else
	{
		if (SPIFFS.exists(fileBak))
		{
			SPIFFS.rename(fileBak, fileCsv);
		}
		else
		{
			//csvファイルもbakファイルも存在しない → csvファイルを生成する
			File file = SPIFFS.open(fileCsv, FILE_WRITE, true);
			if (file) { file.close(); }
		}
	}
}

//CSVファイルからレコードを読み込む
std::vector<String> Record::readCsv(void)
{
	std::vector<String> lines;
	File file = SPIFFS.open(fileCsv, FILE_READ);
	if (!file) { return lines; }
Serial.println("read csv file:");
	while (0 < file.available())
	{
		String line = file.readStringUntil('\n');
		line.trim();	//改行コードを削除する
		if (line.length() == 0) { continue; }
Serial.println(line);
		lines.push_back(line);
	}
	file.close();

	return lines;
}

//ファイルの内容を更新する
bool	Record::update(const String& record)
{
	//CSVファイルからレコードを読み込む
	std::vector<String> lines = readCsv();

	//引数のレコードを追加する
	lines.push_back(record);	//末尾に追加する
	while (recordCountMax < lines.size()) { lines.erase(lines.begin()); }	//先頭行を削除する

	//読み込み元のファイルをバックアップファイルとしてリネームする
	if (SPIFFS.exists(fileBak) && !SPIFFS.remove(fileBak)) { return false; }
	if (!SPIFFS.rename(fileCsv, fileBak)) { return false; }

	//新規ファイルにレコードを保存する
	File file = SPIFFS.open(fileCsv, FILE_WRITE, true);
	if (!file) { return false; }
Serial.printf("new file open. [%s]:%d byte\n", fileCsv, file.size());
	for (uint8_t i = 0; i < lines.size(); i++) { file.println(lines[i]); }
	file.close();
	
	return true;
}

//ファイルを削除する
void	Record::deleteFile(void)
{
	if (SPIFFS.exists(fileUpdateTime)) { SPIFFS.remove(fileUpdateTime); }
	if (SPIFFS.exists(fileCsv)) { SPIFFS.remove(fileCsv); }
	if (SPIFFS.exists(fileBak)) { SPIFFS.remove(fileBak); }
}

//グラフ描画形式でデータを読み込む
void	Record::loadGraphData(void)
{
	//CSVファイルからレコードを読み込む
	std::vector<String> lines = readCsv();

	//レコードからセンサー値を切り出す
	std::vector<String> lineParts;
	std::vector<std::vector<String>> fields;	//全行全列のデータ一つ一つ（DBのフィールドにあたる）
	for (uint8_t i = 0; i < lines.size(); i++)
	{
		lineParts = StrUty::split(lines[i], ',');
		for (uint8_t p = 0; p < lineParts.size(); p++) { lineParts[p].replace("\"", ""); }
		fields.push_back(lineParts);
	}

	//センサー値データを作成する
	buildGraphDataset(fields, graphData.temperature, fieldIdxT);
	buildGraphDataset(fields, graphData.humidity, fieldIdxH);
	buildGraphDataset(fields, graphData.pressure, fieldIdxP);

	//日時データを作成する
	buildGraphLabel(fields, graphData.dateTime);
}

//GraphDataオブジェクトに値をセットする
void	Record::buildGraphDataset(const std::vector<std::vector<String>>& fields, GraphData& graphData, const uint8_t fieldIdx)
{
	graphData.initialize();

	for (uint8_t i = 0; i < fields.size(); i++)
	{
		float num = fields[i][fieldIdx].toFloat();
		graphData.max = std::max(graphData.max, num);
		graphData.min = std::min(graphData.min, num);
		graphData.jointData += fields[i][fieldIdx] + ',';
	}

	graphData.scaleMax = (int16_t)std::ceil(graphData.max);		//切り上げ
	graphData.scaleMin = (int16_t)std::floor(graphData.min);	//切り捨て

	const std::vector<String>& lastRecord = fields[fields.size() - 1];
	graphData.lastData = String(lastRecord[fieldIdx]);
}

//GraphDataオブジェクトにラベル（日時）をセットする
void	Record::buildGraphLabel(const std::vector<std::vector<String>>& fields, GraphData& graphData)
{
	graphData.initialize();
	for (uint8_t i = 0; i < fields.size(); i++)
	{
//		graphData.jointData += "'" + fields[i][fieldIdxYMD] + ' ' + fields[i][fieldIdxHM] + "',";
		graphData.jointData += "'" + fields[i][fieldIdxHM] + "',";
	}

	const std::vector<String>& lastRecord = fields[fields.size() - 1];
	graphData.lastData = lastRecord[fieldIdxYMD] + ' ' + lastRecord[fieldIdxHM];
}

//グラフのHTMLファイルを保存する
bool	Record::saveGraphHtmlFile(String& graphHtml)
{
	return writeFile(fileGraphHtml, graphHtml);
}

//グラフのHTMLのダウンロード用のファイル名を作成する
String	Record::graphDownloadName(void)
{
	String datetime = graphData.dateTime.lastData;	//"YYYY/M/D HH:MM"
	datetime.replace('/', '-');
	datetime.replace(':', '-');
	datetime.replace(' ', '_');	//"YYYY-M-D_HH-MM"
	String downloadName = String(fileGraphHtml);
	downloadName.replace(".", "_" + datetime + ".");
	return downloadName;
}
