
#include <Arduino.h>
#include <WiFi.h>
#include <WebServer.h>
#include "NvsWiFiSetting.hpp"
#include "WiFiApServer.hpp"
#include "WiFiApServerHtml.hpp"
#include "Record.hpp"

WebServer	webServer;
extern	Record	record;		//main()に実体あり

//webページを出力する関数
void	webSendHtml(int responseCode, const String& body);
void	hNotFound(void);
void	hRoot(void);
void	hInputParam(void);
void	hUpdateParam(void);
void	hEraseNvs(void);
void	hRestartProduct(void);
void	hDisplaySensorValue(void);

//APモードでWebサーバーを起動する
IPAddress	WiFiApServer::serverSetup(void)
{
//	WiFi.mode(WIFI_MODE_APSTA);	//APモード＋STAモード …である必要はない
	WiFi.mode(WIFI_MODE_AP);
	WiFi.softAP(apSsid, apPswd);
	webServer.serveStatic("/chart.umd.js", SPIFFS, "/chart.umd.js");
	webServer.serveStatic(record.csvFilename(), SPIFFS, record.csvFilename());
	webServer.on("/", HTTP_GET, hRoot);	//<a href> -> GET:/i/?scan=yes-no {and} GET:/e/?erase=yes
	webServer.on("/i/", HTTP_GET, hInputParam);	//<form> -> POST:/i/
	webServer.on("/i/", HTTP_POST, hUpdateParam);	//update:ssid,pswd; <a href> -> root {and} <form>POST:/r/
	webServer.on("/e/", HTTP_GET, hEraseNvs);	//NVS_Erase(); <a href> -> root
	webServer.on("/r/", HTTP_POST, hRestartProduct);	//ESP.restart();
	webServer.on("/sv/", HTTP_GET, hDisplaySensorValue);	//Content-Type:text/plain
	webServer.onNotFound(hNotFound);
	webServer.begin();
//Serial.println("WebServer(AP) IP-address: " + WiFi.softAPIP().toString());
	return WiFi.softAPIP();		//192.168.4.1	← WiFi.softAP()のデフォルト値
}

// クライアント接続の有無をポーリングする
void	WiFiApServer::listening(void)
{
	webServer.handleClient();
}

//指定されたbody部をはめ込んだHTMLドキュメントをサーバーから出力する
//引数	responseCode: 200 or 404
//		body: htmlの<body>...</body>の中の部分
void	webSendHtml(int responseCode, const String& body)
{
	String html = htmlDocTemplate;
	html.replace("{BODY}", body);
	webServer.send(responseCode, "text/html", html);
}

//ハンドル：Not Found
void	hNotFound(void)
{
	String body = "<h1>404 Not Found</h1>\n<p>URI: " + webServer.uri() + "</p>\n";
	webSendHtml(404, body);
}

// ハンドル：Root/
void	hRoot(void)
{
	String body = rootBody;
	body.replace("{FILE_CSV}", RECORD_FILENAME_CSV);
	webSendHtml(200, body);
}

//ハンドル：Root/i/ (GET)
void	hInputParam(void)
{
	if (webServer.arg("scan").equals("yes"))
	{
		//<form>のSSIDパートを<select/option>とする場合
		//周辺のSSIDをスキャンする（開発環境では8秒程度かかった）
//Serial.println("scan SSID ...");
		int16_t result = WiFi.scanNetworks();	//0以上の個数か、負数のエラーコードが返る
		uint8_t ssidCount = static_cast<uint8_t>(result);	//負数やUINT8_MAXより大きい値の状況は無視

		//SSIDリスト(htmlの<optiopn>)を作成する
		String optionList = "";
		for (uint8_t i = 0; i < ssidCount; i++)
		{
			String ssidName = WiFi.SSID(i);
			optionList += "<option value=\"" + ssidName + "\">" + ssidName + "</option>\n";
//Serial.println(ssidName);
		}

		//<form>の<select>に<optiopn>リストをはめ込む
		inputParamSelect.replace("{OPTION}", optionList);
		inputParamBody.replace("{SSID_PART}", inputParamSelect);
	}
	else
	{
		//<form>のSSIDパートを<input>とする場合
		inputParamBody.replace("{SSID_PART}", inputParamInput);
	}
	
	webSendHtml(200, inputParamBody);
}

//ハンドル：Root/i/ (POST)
void	hUpdateParam(void)
{
	String userSsid = webServer.arg("userssid");
	String userPswd = webServer.arg("userkey");
//Serial.printf("hUpdateWiFiParam()-[%s][%s]\n", userSsid.c_str(), userPswd.c_str());

	//NVSのSTAのssid/pswdを書き換える
	//・STAモードで接続開始した時点でNVSにsta.ssid,sta.pswdが保存される。
	NvsWiFiSetting nvs;
	nvs.writeWifiParam(userSsid.c_str(), userPswd.c_str());
//nvs.debugPrint("sta.ssid/sta.pswd update");

	webSendHtml(200, updateParamBody);
}

//ハンドル：Root/e/ (GET)
void	hEraseNvs(void)
{
	if (webServer.arg("erase").equals("yes"))
	{
		//NVS全域を消去する（STAのssid/pswdを消去する目的）
		NvsWiFiSetting nvs;
		nvs.eraseNvs();
//nvs.debugPrint("sta.ssid/sta.pswd erase");

		webSendHtml(200, dataEraseNvsBody);
	}
	else
	{
		hNotFound();
	}
}

//ハンドル：Root/r/ (POST)
void	hRestartProduct(void)
{
	//ESP32を再起動する
	if (!webServer.arg("restart").isEmpty()) { ESP.restart(); }
}

//ハンドル：Root/sv/ (GET)
void	hDisplaySensorValue(void)
{
	String sensorValue = webServer.arg("sval");
	if (sensorValue.equals("graph"))
	{
		//グラフを出力する
		//・データを整形してchartBodyに埋め込んでサーバーから出力する。
		record.loadGraphData();
		GraphDataGroup& graph = record.graphData;
		String body = chartBody;

		//最新情報欄
		body.replace("{LAST_DATETIME}", graph.dateTime.lastData);
		body.replace("{LAST_TEMPERATURE}", graph.temperature.lastData);
		body.replace("{LAST_HUMIDITY}", graph.humidity.lastData);
		body.replace("{LAST_PRESSURE}", graph.pressure.lastData);

		//グラフ部分
		body.replace("{LABEL_DATETIME}", graph.dateTime.jointData);
		body.replace("{DATA_TEMPERATURE}", graph.temperature.jointData);
		body.replace("{DATA_HUMIDITY}", graph.humidity.jointData);
		body.replace("{DATA_PRESSURE}", graph.pressure.jointData);
		body.replace("{SCALE_TEMPERATURE_MAX}", String(graph.temperature.scaleMax));
		body.replace("{SCALE_TEMPERATURE_MIN}", String(graph.temperature.scaleMin));
		body.replace("{SCALE_HUMIDITY_MAX}", String(graph.humidity.scaleMax));
		body.replace("{SCALE_HUMIDITY_MIN}", String(graph.humidity.scaleMin));
		body.replace("{SCALE_PRESSURE_MAX}", String(graph.pressure.scaleMax));
		body.replace("{SCALE_PRESSURE_MIN}", String(graph.pressure.scaleMin));

		webSendHtml(200, body);
	}
	else if (sensorValue.equals("delete"))
	{
		//センサー値の記録ファイルを削除する
		record.deleteFile();
		webSendHtml(200, dataDeleteBody);
	}
	else
	{
		//不正なパラメータだった場合
		hNotFound();
	}
}
