#include <Arduino.h>
#include <Wire.h>	//I2C
#include "WiFiWps.hpp"
#include "WiFiApServer.hpp"
#include "Display.hpp"
#include "Bme280.hpp"	//温度・湿度・気圧センサー
//#include "BH1750.hpp"	//照度センサー
#include "TSL2561.hpp"	//照度センサー
#include "Record.hpp"

#define	ONESEC_IN_MICRO	1000000

WiFiApServer	webApp;
WiFiWps		wifi;
struct tm	timeInfo;
Display	vfd;
Bme280	sensor;
TBme280Value	tValue;
//BH1750	illumiSensor;
TSL2561	illumiSensor;
Record	record;

//変数
const	uint8_t	i2cSda = GPIO_NUM_2,	i2cScl = GPIO_NUM_8;
const	uint8_t	buttonA = GPIO_NUM_9;
int8_t	countDownSec = 0;		//ボタン押下時の日付表示のカウントダウンタイマー（秒）
int		lastAdjustTime = -1;	//struct_tm.tm_hourの値(0-23)、または -1（エラー発生時）
uint16_t	lux;	//照度センサーから取得した明るさ

//変数：debug用
//const	uint8_t	ledPin = GPIO_NUM_1;
//static	uint8_t	ledSwitch = HIGH;

//関数
static	void	programFinish(const char* msg);
static	void	setupDisplay(void);
static	bool	setupWifi(void);
static	bool	setupNtp(void);
static	bool	adjustTime(void);
static	void	startWebServer(void);
static	void	clockSecondIncrement(struct tm* tmInfo, uint8_t sec = 1);
static	bool	isAdjustTime(const struct tm* tmInfo);
static	bool	isReadSensor(const struct tm* tmInfo);
static	bool	isWriteSensor(const struct tm* tmInfo);

void	setup(void)
{
//	Serial.begin(115200);

	/*	USB_D+/D-を無効化する
	・開発環境ではないPCに当デバイスを接続した場合に当デバイスを見えないようにするため、
		GPIO18,19(USB_D+/D-)を論理的に未接続状態にする。
	・その状態では当然プログラムは書き込めないが、GPIO9をGNDに接続（当デバイスのボタンAを押下）して起動すれば、
		（SPIブートモードではなく）Downloadモードでの起動となり、プログラムは書き込める。
	*/
	pinMode(GPIO_NUM_18, INPUT);	//INPUT: Hi-Z
	pinMode(GPIO_NUM_19, INPUT);	//INPUT: Hi-Z

	//初期化
	Wire.begin(i2cSda, i2cScl);
	vfd.initialize();
	sensor.initialize();
	illumiSensor.Initialize();

	//入力ボタンを設定する
	pinMode(buttonA, INPUT_PULLUP);
//	pinMode(ledPin, OUTPUT);	//debug

	//VFD表示安定待ち
	setupDisplay();

	//WiFi接続する
	//・NVSを消去した場合はNGが表示される。
	bool isOK = setupWifi();

	//NTPサーバーを指定し（日本標準時）、時刻を取得する
	//・WiFi通信できない場合はNGが表示される。
	isOK = setupNtp();

	//Webサーバーを起動する
	//・Webサーバーは常時起動した状態。
	//・本体の初回起動時、および、NVSを消去した場合、
	//	ユーザーはWebサーバに接続し、WiFiパスワードを設定する必要がある。
	startWebServer();

	//環境センサーの値を取得する
	sensor.getSensorValue(&tValue);	//構造体に値をセットする
	lux = illumiSensor.Lux();
}

void	loop(void)
{
	//1ループの経過時間を測定開始する
	uint64_t startTime = micros();

//	digitalWrite(ledPin, ledSwitch = (HIGH + LOW) - ledSwitch);	//debug

	//ボタンが押された場合、または、それによる日付表示のカウントダウン中の場合
	bool isButtonOn = (digitalRead(buttonA) == LOW);
	if (isButtonOn || 0 < countDownSec)
	{
		//日付表示を開始する
		if (isButtonOn)
		{
			countDownSec = 3;
			vfd.displayDate(&timeInfo);
		}
		else
		{
			countDownSec--;
		}
	}
	else
	{
		//現在の時刻（秒）に応じて時刻やセンサー値を表示する
		vfd.displayTimeAndSensorValue(timeInfo, tValue, lux);
	}

	//時刻合わせのタイミングだったらNTPで時刻合わせをする
	if (isAdjustTime(&timeInfo))
	{
		getLocalTime(&timeInfo);
		lastAdjustTime = timeInfo.tm_hour;	//今回の時刻を保存する
	}

	//センサー値を取得するタイミングだったらセンサーからデータを取得する
	if (isReadSensor(&timeInfo)) { sensor.getSensorValue(&tValue);	lux = illumiSensor.Lux(); }

	//センサー値を保存するタイミングだったら保存する
	if (isWriteSensor(&timeInfo)) { record.write(timeInfo, tValue); }

	//WebServerへのクライアント接続をポーリングする
	webApp.listening();

	//今回のloop()にかかった時間
	uint64_t elapseTime = micros() - startTime;
//Serial.printf("elapseTime = %dus\n", elapseTime);

	//loop()が1秒を超えた場合、その秒数（秒単位）だけ時計を進める
	if (ONESEC_IN_MICRO < elapseTime)
	{
//Serial.printf("1sec < elapseTime = %dus\n", elapseTime);
		uint8_t sec = elapseTime / ONESEC_IN_MICRO;
		clockSecondIncrement(&timeInfo, sec);
		elapseTime -= ONESEC_IN_MICRO * sec;
	}

	//1秒経過の時間調整
	uint64_t sleepTime = ONESEC_IN_MICRO - elapseTime;
	delayMicroseconds(sleepTime);

	//時刻を1秒進める
	clockSecondIncrement(&timeInfo);
}

//プログラム停止のための無限ループ
static	void	programFinish(const char* msg)
{
//Serial.printf("finished. - %s\n", msg);
	while (1) {}	
}

//VFD表示安定待ち（フィラメント温め）
static	void	setupDisplay(void)
{
	for (uint8_t i = 1; i <= vfd.gridMax(); i++)
	{
		vfd.displayString(i, String(i));
		delay(300);	//ms
	}
	vfd.clear();
}

//WiFi接続する
static	bool	setupWifi(void)
{
	uint8_t	grid = vfd.displayString(1, "WiFi ");
	bool isOK = wifi.wpsSetup();
	vfd.displayString(grid, isOK ? "OK" : "NG");
	delay(1000);
	vfd.clear();
	return isOK;
}

//NTPサーバーを指定し（日本標準時）、時刻を取得する
static	bool	setupNtp(void)
{
	uint8_t grid = vfd.displayString(1, "ntp ");
	configTime(3600 * 9, 0, "ntp.nict.jp", "time.google.com");
	bool isOK = adjustTime();
	vfd.displayString(grid, isOK ? "OK" : "NG");
	delay(1000);
	vfd.clear();
	return isOK;
}

//時刻合わせをする（NTP通信）
static	bool	adjustTime(void)
{
	bool isOK = getLocalTime(&timeInfo);
	lastAdjustTime = isOK ? timeInfo.tm_hour : -1;
	return isOK;
}

//WebServerを起動する
static	void	startWebServer(void)
{
	IPAddress ipAddr = webApp.serverSetup();
	vfd.clear();
	vfd.displayIPAddress(ipAddr);
	delay(3000);

	vfd.clear();
	vfd.displayString(1, webApp.pswd());
	delay(3000);

	vfd.clear();
}

//時刻をn秒進める
static	void	clockSecondIncrement(struct tm* tmInfo, uint8_t sec)
{
	time_t mkt = mktime(tmInfo);
	mkt += sec;
	struct tm* newTmInfo = localtime(&mkt);
	*tmInfo = *newTmInfo;	//構造体の代入（ポインタの代入ではなく）
}

//時刻合わせするタイミングか否か
static	bool	isAdjustTime(const struct tm* tmInfo)
{
	//1時00分00秒から3時間ごとに時刻合わせする（参考：半日で1分程度ずれる）
	//・判定時刻から5分ごとに、時刻合わせしたかを確認する。
	//	→時刻表示モードなど長時間処理によって判定時刻をすり抜けている可能性があるので。
	return	(tmInfo->tm_hour % 3 == 1) &&
			(tmInfo->tm_min % 5 == 0) &&
			(tmInfo->tm_sec == 0) &&
			(lastAdjustTime != tmInfo->tm_hour);
}

//センサーを読み取るタイミングか否か
static	bool	isReadSensor(const struct tm* tmInfo)
{
	//xx時xx分30秒に取得する
	//※時刻表示モードなど長時間処理によって当処理の時刻判定をすり抜けても気にしなくてよい
	//→短時間の周期（毎分）で取得するので。
	return	(tmInfo->tm_sec == 30);
}

//センサー値を保存するタイミングか否か
static	bool	isWriteSensor(const struct tm* tmInfo)
{
	//xx時00分00秒に保存する
	//・判定時刻から5分ごとに保存したかを確認する。
	//	→時刻表示モードなど長時間処理によって当処理の時刻判定をすり抜ける可能性があるので。
	return	(tmInfo->tm_min % 5 == 0) &&
			(tmInfo->tm_sec == 0) &&
			!record.isUpdatedThisHour(*tmInfo);
}
